from types import ModuleType
from typing import Union

import pytest
from pytest_lambda.impl import LambdaFixture

__all__ = ['precondition_fixture']


def precondition_fixture(*args, **kwargs):
    """Fixture to be evaluated before the Common Subject is invoked

    NOTE: this fixture only makes sense when using CommonSubjectTestMixin, or
          its subclasses.

    Usage:

        class TestMyStuff(CommonSubjectTestMixin):
            common_subject = static_fixture(datetime.utcnow)
            before = precondition_fixture(lambda: datetime.utcnow())

            def it_evaluates_precondition_before_subject(self, common_subject_rval, before):
                assert before < common_subject_rval

    """
    return PreconditionFixture(*args, **kwargs)


class PreconditionFixture(LambdaFixture):
    def contribute_to_parent(self, parent: Union[type, ModuleType], name: str, **kwargs):
        # Add precondition marker to class. CommonSubjectTestMixin.marked_preconditions
        # will load all fixtures named by this marker.
        pytest.mark.precondition(name)(parent)

        return super().contribute_to_parent(parent, name, **kwargs)



from typing import Callable

import pytest
from pytest_lambda import lambda_fixture, static_fixture

from pytest_common_subject import (
    CommonSubjectTestMixin,
    DeferredCommonSubjectRvalUsage,
    precondition_fixture,
    WithCommonSubjectDeferred,
)


class DescribeCommonSubjectTestMixin:

    @pytest.fixture
    def get_next_val(self) -> Callable[[], int]:
        """Method returning a higher number each invocation per test
        """
        counter = 0

        def get_next_val() -> int:
            nonlocal counter
            counter += 1
            return counter

        return get_next_val


    class TestIsCommonSubjectDeferred(CommonSubjectTestMixin):
        common_subject = lambda_fixture('get_next_val')

        class AssertNotDeferred:
            """Test context mixin verifying traits of an undeferred common_subject
            """
            def it_calls_common_subject_automatically(self, call_common_subject):
                expected = 2  # it ought to have already been called once,
                              # so we get value #2 when calling it below
                actual = call_common_subject()
                assert expected == actual

            def it_allows_usage_of_common_subject_rval(self, common_subject_rval):
                assert common_subject_rval == common_subject_rval


        class AssertDeferred:
            """Test context mixin verifying traits of n deferred common_subject
            """
            def it_does_not_call_common_subject_automatically(self, call_common_subject):
                expected = 1  # it ought not to have been called,
                              # so we get value #1 when calling it below
                actual = call_common_subject()
                assert expected == actual

            def it_disallows_usage_of_common_subject_rval(self, common_subject_rval):
                with pytest.raises(DeferredCommonSubjectRvalUsage):
                    assert common_subject_rval == common_subject_rval


        class ContextNotDeferred(AssertNotDeferred):
            is_common_subject_deferred = static_fixture(False)


        class ContextDeferred(AssertDeferred):
            is_common_subject_deferred = static_fixture(True)


        class ContextDeferredByMixin(
            WithCommonSubjectDeferred,
            AssertDeferred,
        ):
            pass


    class TestPreconditions(CommonSubjectTestMixin):

        # This essentially causes the common_subject_rval fixture to invoke this method
        common_subject = lambda_fixture('get_next_val')


        class TestOverridingPreconditionsFixture:
            # Grab a val at the beginning
            first_val = lambda_fixture(lambda get_next_val: get_next_val())

            # Grab a val late (but should still be before common_subject_rval!)
            late_val = pytest.mark.late(lambda_fixture(lambda get_next_val: get_next_val()))

            preconditions = lambda_fixture('first_val', 'late_val')

            def it_evaluates_preconditions_before_invoking_subject(self, common_subject_rval, first_val, late_val):
                assert first_val < late_val < common_subject_rval


        class TestPreconditionMark:
            # Grab a val at the beginning
            first_val = lambda_fixture(lambda get_next_val: get_next_val())

            # Grab a val late (but should still be before common_subject_rval!)
            late_val = pytest.mark.late(lambda_fixture(lambda get_next_val: get_next_val()))

            @pytest.mark.precondition('first_val')
            @pytest.mark.precondition('late_val')
            def it_evaluates_preconditions_before_invoking_subject(self, common_subject_rval, first_val, late_val):
                assert first_val < late_val < common_subject_rval


        class TestPreconditionFixture:
            # Grab a val at the beginning
            first_val = precondition_fixture(lambda get_next_val: get_next_val())

            # Grab a val late (but should still be before common_subject_rval!)
            late_val = pytest.mark.late(precondition_fixture(lambda get_next_val: get_next_val()))

            def it_evaluates_preconditions_before_invoking_subject(self, common_subject_rval, first_val, late_val):
                assert first_val < late_val < common_subject_rval



import pkg_resources

from .exceptions import DeferredCommonSubjectRvalUsage
from .fixtures import precondition_fixture
from .mixins import CommonSubjectTestMixin, WithCommonSubjectDeferred

__version__ = pkg_resources.get_distribution('pytest-common-subject').version

__all__ = [
    'DeferredCommonSubjectRvalUsage',
    'precondition_fixture',
    'CommonSubjectTestMixin',
    'WithCommonSubjectDeferred',
]



from typing import Type

from lazy_object_proxy import Proxy


class VolatileValue(Proxy):
    """Special object which raises on any attempt to utilize it

    An instance of this class is returned from the `common_subject_rval`
    fixture when `is_common_subject_deferred` is set to True. This is to warn
    the user early of unintended usages, which could possibly lead to false
    negatives (passing tests that should be failing).
    """

    def __init__(self,
                 message: str = 'An attempt to use a VolatileValue was made',
                 exc_class: Type[Exception] = RuntimeError):

        def factory():
            raise exc_class(message)

        super().__init__(factory)




from typing import Callable, Any, TypeVar, Dict

import pytest

from .exceptions import DeferredCommonSubjectRvalUsage
from .util import VolatileValue

__all__ = [
    'CommonSubjectTestMixin',
    'WithCommonSubjectDeferred',
]

V = TypeVar('V')


class CommonSubjectTestMixin:
    """Abstract mixin for test contexts revolving around a common target

    Imperative tests are intensive to read. To grok what a test is *actually*
    trying to prove requires inspecting many lines of code which are ultimately
    inessential boilerplate.

    If, instead, we group our tests by the Subject being prodded, we can cordon
    off boilerplate to a single location, leaving only the juicy bits left to
    read.

    This mixin will call a single function for every test method, and provide
    its return value as a fixture, which is always evaluated after all other
    fixtures (so they can still be used for setup). The args passed in this
    function call can be customized by overriding the args/kwargs fixtures.

    Test contexts defined underneath this mixin are free to override
    args/kwargs, and create/override fixtures to setup the environment for the
    function call. This encourages named, well-defined contexts for testing
    the common subject from different viewpoints.

    This pattern also encourages slimmer tests, ideally responsible for testing
    a single aspect of the common subject and named descriptively. The upshot
    of this is all issues can be reported in one test run – as opposed to
    monolith tests, where fixing an issue and re-running the test suite reveals
    the next thing to fix.
    """

    @pytest.fixture
    def common_subject(self, *args) -> Callable[..., V]:
        """The method being tested in this context"""
        raise NotImplementedError(
            'Please override the `common_subject` fixture and provide a callable.')

    @pytest.fixture
    def is_common_subject_deferred(self) -> bool:
        """Whether to hold off on calling common_subject automatically

        If True, common_subject is not called in the `common_subject_rval`
        fixture. Instead, `common_subject_rval` will return a special DEFERRED
        value, which will raise an exception if an attempt is made to use the
        value — this is to aid in preventing unintended misuse.

        Deferring `common_subject_rval` is useful when one wishes to test
        whether `common_subject` raises an exception.

        Example:

            class DescribeMyErroneousMethod(
                CommonSubjectTestMixin
            ):
                common_subject = static_fixture(my_erroneous_method)
                is_common_subject_deferred = static_fixture(True)

                def it_raises_exception(self, call_common_subject):
                    with pytest.raises(MyException):
                        call_common_subject()

        See also the WithCommonSubjectDeferred test mixin
        """
        return False

    @pytest.fixture(autouse=True)
    @pytest.mark.late  # this ensures the fixture is executed at end of setup
                       # NOTE: this MUST be placed after pytest.fixture
    def common_subject_rval(self,
                            is_common_subject_deferred: bool,
                            call_common_subject: Callable[[], V],
                            all_preconditions: Any,
                            ) -> V:
        """The return value from invoking the common subject in this env

        To perform any post-processing on the return value (like parsing the
        JSON of an HTTP response), it's recommended to define another fixture
        that requests `common_subject_rval`.
        """
        if is_common_subject_deferred:
            return VolatileValue(
                'An attempt was made to use `common_subject_rval` while '
                '`is_common_subject_deferred` is True. This exception is '
                'thrown to prevent accidental passing of tests through '
                'unintended use of a deferred `common_subject_rval`',
                exc_class=DeferredCommonSubjectRvalUsage,
            )

        return call_common_subject()

    @pytest.fixture
    def call_common_subject(self,
                            common_subject: Callable[..., V],
                            args: tuple,
                            kwargs: Dict[str, Any],
                            ) -> Callable[[], V]:
        """A 0-arg method which invokes common_subject and returns its result

        Override this to customize how the common subject is invoked.
        """
        def call_common_subject() -> V:
            return common_subject(*args, **kwargs)
        return call_common_subject

    @pytest.fixture
    def args(self) -> tuple:
        """Inline arguments to be passed when invoking common_subject"""
        return ()

    @pytest.fixture
    def kwargs(self) -> dict:
        """Keyword arguments to be passed when invoking common_subject"""
        return {}

    @pytest.fixture
    def all_preconditions(self, preconditions: Any, marked_preconditions: Any):
        """Metafixture to request preconditions declared by any means

        Preconditions can be declared by overriding the `preconditions`
        fixture, using the @pytest.mark.precondition('fixture_name') marker,
        or by using the `precondition_fixture()` lambda fixture declaration.
        `all_preconditions` loads each of these types of preconditions.

        `common_subject_rval` requests `all_preconditions`, instead of directly requesting
        `preconditions` and `marked_preconditions`, so the user may customize
        precondition loading without having to completely redefine `common_subject_rval`.
        """

    @pytest.fixture
    def preconditions(self):
        """Any preconditions to be evaluated before the subject is invoked

        This fixture is requested by the common_subject_rval fixture, before calling
        call_common_subject(). By overriding this fixture, one can request
        other fixtures to be evaluated as "preconditions".

        Example:

            class TestMyStuff(CommonSubjectTestMixin):
                common_subject = lambda_fixture(lambda: MyModel.objects.all().delete())

                initial_count = lambda_fixture(lambda: MyModel.objects.count())
                preconditions = lambda_fixture('initial_count')

                def it_deletes_things(self, initial_count):
                    assert initial_count == 1
                    assert MyModel.objects.count() == 0

        """

    @pytest.fixture
    def marked_preconditions(self, request):
        """Load any fixtures declared with @pytest.mark.precondition('name')

        NOTE: this is how precondition_fixture() declares preconditions
        """
        precondition_marks = tuple(request.node.iter_markers('precondition'))
        precondition_fixtures = [mark.args[0] for mark in precondition_marks]

        for fixturename in precondition_fixtures:
            request.getfixturevalue(fixturename)


class WithCommonSubjectDeferred:
    """Test context mixin which sets `is_common_subject_deferred` to True

    This will defer the automatic calling of `common_subject`, leaving its
    invocation the responsibility of the user.
    """

    @pytest.fixture
    def is_common_subject_deferred(self) -> bool:
        return True
